/*************************************************************************/
/*                                                                       */
/*  Copyright (c) 1994 Stanford University                               */
/*                                                                       */
/*  All rights reserved.                                                 */
/*                                                                       */
/*  Permission is given to use, copy, and modify this software for any   */
/*  non-commercial purpose as long as this copyright notice is not       */
/*  removed.  All other uses, including redistribution in whole or in    */
/*  part, are forbidden without prior written permission.                */
/*                                                                       */
/*  This software is provided with absolutely no warranty and no         */
/*  support.                                                             */
/*                                                                       */
/* --------------------------------------------------------------------- */
/*                                                                       */
/*  Modifications of the original Barnes-Hut code (as taken from         */
/*  Stanford's SPLASH-2 distribution) to allow use on Alewife and        */
/*  with CRL are copyright:                                              */
/*                                                                       */
/*  Copyright (C) 1995 Massachusetts Institute of Technology             */
/*                                                                       */
/*************************************************************************/

/* only run on processor 0!!! */

#include "top.h"

#include "math.h"
#include "stdio.h"
#include "mdvar.h"
#include "water.h"
#include "cnst.h"
#include "fileio.h"
#include "parameters.h"
#include "mddata.h"
#include "split.h"

INITIA()
{
    /*   this routine initializes the positions of the molecules along
         a regular cubical lattice, and randomizes the initial velocities of
         the atoms.  The random numbers used in the initialization of velocities 
         are read from the file random.in, which must be in the current working
         directory in which the program is run  */
    
    static double XMIN = 0;
    static double YMIN = 0;
    static double ZMIN = 0;
    FILE *random_numbers;       /* points to input file containing
                                   pseudo-random numbers for initializing
                                   velocities */
    double XMAS[4], XS, ZERO, WCOS, WSIN, XT[4], YT[4], Z;
    double SUX, SUY, SUZ, SUMX, SUMY, SUMZ, FAC;
    int mol=0;
    int atom=0;
    int deriv;
    
    random_numbers = fopen(RANDOM_FILE, "r");
    if (random_numbers == NULL)
    {
      fprintf(stderr, "%s:%d unable to open RANDOM_FILE (%s)\n",
	      __FILE__, __LINE__, RANDOM_FILE);
      fflush(stderr);
      exit(-1);
    }
    
    XMAS[1]=sqrt(OMAS*HMAS);
    XMAS[0]=HMAS;
    XMAS[2]=HMAS;
    
    /* .....assign positions */
    {
/*        double NS = pow((double) NMOL, 1.0/3.0) - 0.00001;*/
        double NS;

	for(NS=1; NS<17; NS++) {
	    if (pow(NS, 3.0) == NMOL) break;
	}
	{
        double XS = BOXL/NS;
        double ZERO = XS * 0.50;
        double WCOS = ROH * cos(ANGLE * 0.5);
        double WSIN = ROH * sin(ANGLE * 0.5);
        int i,j,k;
        
        printf("\nNS = %.16f\n",NS);
        printf("BOXL = %10f\n",BOXL);
        printf("CUTOFF = %10f\n",CUTOFF);
        printf("XS = %10f\n",XS);
        printf("ZERO = %g\n",ZERO);
        printf("WCOS = %f\n",WCOS);
        printf("WSIN = %f\n",WSIN);
        fflush(stdout);
        
#if defined(RANDOM)
	assert(0);
#else
        /* not random initial placement, but rather along a regular
           lattice.  This is the default and the prefered initialization
           since random does not necessarily make sense from the viewpoint
           of preserving bond distances */
        
        fprintf(six, "***** NEW RUN STARTING FROM REGULAR LATTICE *****\n");
        fflush(six);
        XT[2] = ZERO;
        mol = 0;
        for (i=0; i < NS; i+=1) {
            XT[1]=XT[2]+WCOS;
            XT[3]=XT[1];
            YT[2]=ZERO;
            for (j=0; j < NS; j+=1) {
                YT[1]=YT[2]+WSIN;
                YT[3]=YT[2]-WSIN;
                Z=ZERO;
                for (k = 0; k < NS; k++) {

#if defined(USE_CRL)
		    /*VAR_g_mol = (molecule_type *) rgn_map(VAR_rids[mol]);*/
		    VAR_g_mol = VAR[mol];
		    rgn_start_write(VAR_g_mol);
#else
		    VAR_g_mol = VAR[mol];
#endif

                    for (atom = 0; atom < NATOMS; atom +=1) {
                        VAR_g_mol->F[DISP][XDIR][atom] = XT[atom+1];
                        VAR_g_mol->F[DISP][YDIR][atom] = YT[atom+1];
                        VAR_g_mol->F[DISP][ZDIR][atom] = Z;
                    }

#if defined(USE_CRL)
		    rgn_end_write(VAR_g_mol);	
		    /*rgn_unmap(VAR_g_mol);*/
#endif

                    mol += 1;
                    Z=Z+XS;
                }
                YT[2]=YT[2]+XS;
            }
            XT[2]=XT[2]+XS;
        }
        
        if (NMOL != mol) {
            printf("Lattice init error: total mol %d != NMOL %d\n", mol, NMOL);
            exit(-1);
        }
#endif
    }
    }
    
    /* ASSIGN RANDOM MOMENTA */
    fscanf(random_numbers,"%lf",&SUX);
    
    SUMX=0.0;
    SUMY=0.0;
    SUMZ=0.0;
    /*   read pseudo-random numbers from input file random.in */
    for (mol = 0; mol < NMOL; mol++) {

#if defined(USE_CRL)
	/*VAR_g_mol = (molecule_type *) rgn_map(VAR_rids[mol]);*/
	VAR_g_mol = VAR[mol];
	rgn_start_write(VAR_g_mol);
#else
	VAR_g_mol = VAR[mol];
#endif

        for (atom = 0; atom < NATOMS; atom++) {
            fscanf(random_numbers,"%lf",&(VAR_g_mol->F[VEL][XDIR][atom]));
            fscanf(random_numbers,"%lf",&(VAR_g_mol->F[VEL][YDIR][atom]));
            fscanf(random_numbers,"%lf",&(VAR_g_mol->F[VEL][ZDIR][atom]));
            SUMX = SUMX + VAR_g_mol->F[VEL][XDIR][atom];
            SUMY = SUMY + VAR_g_mol->F[VEL][YDIR][atom];
            SUMZ = SUMZ + VAR_g_mol->F[VEL][ZDIR][atom];
            for (deriv = ACC; deriv < MAXODR; deriv++) {
                VAR_g_mol->F[deriv][XDIR][atom] = 0.0;
                VAR_g_mol->F[deriv][YDIR][atom] = 0.0;
		VAR_g_mol->F[deriv][ZDIR][atom] = 0.0;
            }
        } /* atoms */

#if defined(USE_CRL)
	rgn_end_write(VAR_g_mol);	
	/*rgn_unmap(VAR_g_mol);*/
#endif

    } /* molecules */
    
    /* find average momenta per atom */
    SUMX=SUMX/(NATOMS*NMOL);
    SUMY=SUMY/(NATOMS*NMOL);
    SUMZ=SUMZ/(NATOMS*NMOL);
    
    /*  find normalization factor so that <k.e.>=KT/2  */
    SUX=0.0;
    SUY=0.0;
    SUZ=0.0;
    for (mol = 0; mol < NMOL; mol++) {

#if defined(USE_CRL)
	/*VAR_g_mol = (molecule_type *) rgn_map(VAR_rids[mol]);*/
	VAR_g_mol = VAR[mol];
	rgn_start_read(VAR_g_mol);
#else
	VAR_g_mol = VAR[mol];
#endif

        SUX = SUX + (pow( (VAR_g_mol->F[VEL][XDIR][H1] - SUMX),2.0)
                     +pow( (VAR_g_mol->F[VEL][XDIR][H2] - SUMX),2.0))/HMAS
                         +pow( (VAR_g_mol->F[VEL][XDIR][O]  - SUMX),2.0)/OMAS;
        
        SUY = SUY + (pow( (VAR_g_mol->F[VEL][YDIR][H1] - SUMY),2.0)
                     +pow( (VAR_g_mol->F[VEL][YDIR][H2] - SUMY),2.0))/HMAS
                         +pow( (VAR_g_mol->F[VEL][YDIR][O]  - SUMY),2.0)/OMAS;
        
        SUZ = SUZ + (pow( (VAR_g_mol->F[VEL][ZDIR][H1] - SUMZ),2.0)
                     +pow( (VAR_g_mol->F[VEL][ZDIR][H2] - SUMZ),2.0))/HMAS
                         +pow( (VAR_g_mol->F[VEL][ZDIR][O]  - SUMZ),2.0)/OMAS;

#if defined(USE_CRL)
	rgn_end_read(VAR_g_mol);	
	/*rgn_unmap(VAR_g_mol);*/
#endif

    }
    FAC=BOLTZ*TEMP*NATMO/UNITM * pow((UNITT*TSTEP/UNITL),2.0);
    SUX=sqrt(FAC/SUX);
    SUY=sqrt(FAC/SUY);
    SUZ=sqrt(FAC/SUZ);
    
    /* normalize individual velocities so that there are no bulk
       momenta  */
    XMAS[1]=OMAS;
    for (mol = 0; mol < NMOL; mol++) {

#if defined(USE_CRL)
	/*VAR_g_mol = (molecule_type *) rgn_map(VAR_rids[mol]);*/
	VAR_g_mol = VAR[mol];
	rgn_start_write(VAR_g_mol);
#else
	VAR_g_mol = VAR[mol];
#endif

        for (atom = 0; atom < NATOMS; atom++) {
            VAR_g_mol->F[VEL][XDIR][atom] = ( VAR_g_mol->F[VEL][XDIR][atom] -
                                           SUMX) * SUX/XMAS[atom];
            VAR_g_mol->F[VEL][YDIR][atom] = ( VAR_g_mol->F[VEL][YDIR][atom] -
                                           SUMY) * SUY/XMAS[atom];
            VAR_g_mol->F[VEL][ZDIR][atom] = ( VAR_g_mol->F[VEL][ZDIR][atom] -
                                           SUMZ) * SUZ/XMAS[atom];
        } /* for atom */

#if defined(USE_CRL)
	rgn_end_write(VAR_g_mol);	
	/*rgn_unmap(VAR_g_mol);*/
#endif

    } /* for mol */
    
    fclose(random_numbers);
    
} /* end of subroutine INITIA */
